"""
Chef code generator.
"""

import base64
import codecs
import errno
import logging
import os
import os.path
import re
import tarfile

from blueprint import util
from blueprint import walk


def chef(b, relaxed=False):
    """
    Generate Chef code.
    """
    c = Cookbook(b.name, comment=b.DISCLAIMER)

    def source(dirname, filename, gen_content, url):
        """
        Create a cookbook_file and execute resource to fetch and extract
        a source tarball.
        """
        pathname = os.path.join('/tmp', filename)
        if url is not None:
            c.execute('curl -o "{0}" "{1}" || wget -O "{0}" "{1}"'.
                          format(pathname, url),
                      creates=pathname)
        elif gen_content is not None:
            c.file(pathname,
                   gen_content(),
                   owner='root',
                   group='root',
                   mode='0644',
                   backup=False,
                   source=pathname[1:])
        if '.zip' == pathname[-4:]:
            c.execute('{0}'.format(pathname),
                      command='unzip "{0}"'.format(pathname),
                      cwd=dirname)
        else:
            c.execute('{0}'.format(pathname),
                      command='tar xf "{0}"'.format(pathname),
                      cwd=dirname)

    def file(pathname, f):
        """
        Create a cookbook_file resource.
        """
        if 'template' in f:
            logging.warning('file template {0} won\'t appear in generated '
                            'Chef cookbooks'.format(pathname))
            return
        c.directory(os.path.dirname(pathname),
                    group='root',
                    mode='0755',
                    owner='root',
                    recursive=True)
        if '120000' == f['mode'] or '120777' == f['mode']:
            c.link(pathname,
                   group=f['group'],
                   owner=f['owner'],
                   to=f['content'])
            return
        if 'source' in f:
            c.remote_file(pathname,
                          backup=False,
                          group=f['group'],
                          mode=f['mode'][-4:],
                          owner=f['owner'],
                          source=f['source'])
        else:
            content = f['content']
            if 'base64' == f['encoding']:
                content = base64.b64decode(content)
            c.file(pathname,
                   content,
                   backup=False,
                   group=f['group'],
                   mode=f['mode'][-4:],
                   owner=f['owner'],
                   source=pathname[1:])

    def before_packages(manager):
        """
        Create execute resources to configure the package managers.
        """
        packages = b.packages.get(manager, [])
        if 0 == len(packages):
            return
        if 1 == len(packages) and manager in packages:
            return
        if 'apt' == manager:
            c.execute('apt-get -q update')
        elif 'yum' == manager:
            c.execute('yum makecache')

    def package(manager, package, version):
        """
        Create a package resource.
        """
        if manager == package:
            return

        if manager in ('apt', 'yum'):
            if relaxed or version is None:
                c.package(package)
            else:
                c.package(package, version=version)

            # See comments on this section in `puppet` above.
            match = re.match(r'^rubygems(\d+\.\d+(?:\.\d+)?)$', package)
            if match is not None and util.rubygems_update():
                c.execute('/usr/bin/gem{0} install --no-rdoc --no-ri ' # No ,
                          'rubygems-update'.format(match.group(1)))
                c.execute('/usr/bin/ruby{0} ' # No ,
                          '$(PATH=$PATH:/var/lib/gems/{0}/bin ' # No ,
                          'which update_rubygems)"'.format(match.group(1)))

            if 'nodejs' == package:
                c.execute('{ ' # No ,
                          'curl http://npmjs.org/install.sh || ' # No ,
                          'wget -O- http://npmjs.org/install.sh ' # No ,
                          '} | sh',
                          creates='/usr/bin/npm')

        # AWS cfn-init templates may specify RPMs to be installed from URLs,
        # which are specified as versions.
        elif 'rpm' == manager:
            c.rpm_package(package, source=version)

        # All types of gems get to have package resources.
        elif 'rubygems' == manager:
            if relaxed or version is None:
                c.gem_package(package)
            else:
                c.gem_package(package, version=version)
        elif re.search(r'ruby', manager) is not None:
            match = re.match(r'^ruby(?:gems)?(\d+\.\d+(?:\.\d+)?)',
                             manager)
            if relaxed or version is None:
                c.gem_package(package,
                    gem_binary='/usr/bin/gem{0}'.format(match.group(1)))
            else:
                c.gem_package(package,
                    gem_binary='/usr/bin/gem{0}'.format(match.group(1)),
                    version=version)

        # Everything else is an execute resource.
        else:
            c.execute(manager(package, version, relaxed))

    def service(manager, service):
        """
        Create a service resource and subscribe to its dependencies.
        """

        # Transform dependency list into a subscribes attribute.
        # TODO Breaks inlining.
        subscribe = []
        def service_file(m, s, pathname):
            f = b.files[pathname]
            if '120000' == f['mode'] or '120777' == f['mode']:
                subscribe.append('link[{0}]'.format(pathname))
            else:
                subscribe.append('cookbook_file[{0}]'.format(pathname))
        walk.walk_service_files(b, manager, service, service_file=service_file)
        def service_package(m, s, pm, package):
            subscribe.append('package[{0}]'.format(package))
        walk.walk_service_packages(b,
                                   manager,
                                   service,
                                   service_package=service_package)
        def service_source(m, s, dirname):
            subscribe.append('execute[{0}]'.format(b.sources[dirname]))
        walk.walk_service_sources(b,
                                  manager,
                                  service,
                                  service_source=service_source)
        subscribe = util.BareString('resources(' \
            + ', '.join([repr(s) for s in subscribe]) + ')')

        kwargs = {'action': [[':enable', ':start']],
                  'subscribes': [':restart', subscribe]}
        if 'upstart' == manager:
            kwargs['provider'] = util.BareString(
                'Chef::Provider::Service::Upstart')
        c.service(service, **kwargs)

    b.walk(source=source,
           file=file,
           before_packages=before_packages,
           package=package,
           service=service)

    return c


class Cookbook(object):
    """
    A cookbook is a collection of Chef resources plus the files and other
    supporting objects needed to run it.
    """

    def __init__(self, name, comment=None):
        """
        """
        if name is None:
            self.name = 'blueprint-generated-chef-cookbook'
        else:
            self.name = str(name)
        self.comment = comment
        self.resources = []
        self.files = {}

    def add(self, resource):
        """
        Resources must be added in the order they're expected to run.
        Chef does not support managing dependencies explicitly.
        """
        self.resources.append(resource)

    def directory(self, name, **kwargs):
        """
        Create a directory resource.
        """
        self.add(Resource('directory', name, **kwargs))

    def link(self, name, **kwargs):
        """
        Create a (symbolic) link resource.
        """
        self.add(Resource('link', name, **kwargs))

    def file(self, name, content, **kwargs):
        """
        Create a file or cookbook_file resource depending on whether the
        cookbook is dumped to a string or to files.
        """
        self.add(File(name, content, **kwargs))

    def remote_file(self, name, **kwargs):
        """
        Create a remote_file resource.
        """
        self.add(Resource('remote_file', name, **kwargs))

    def package(self, name, **kwargs):
        """
        Create a package resource provided by the default provider.
        """
        self.add(Resource('package', name, **kwargs))

    def rpm_package(self, name, **kwargs):
        """
        Create a package resource provided by RPM.
        """
        self.add(Resource('rpm_package', name, **kwargs))

    def gem_package(self, name, **kwargs):
        """
        Create a package resource provided by RubyGems.
        """
        self.add(Resource('gem_package', name, **kwargs))

    def execute(self, name, **kwargs):
        """
        Create an execute resource.
        """
        self.add(Resource('execute', name, **kwargs))

    def service(self, name, **kwargs):
        """
        Create a service resource.
        """
        self.add(Resource('service', name, **kwargs))

    def _dump(self, w, inline=False):
        """
        Generate Chef code.  This will call the callable `w` with each
        line of output.  `dumps` and `dumpf` use this to append to a list
        and write to a file with the same code.

        If present, a comment is written first.  Next, resources are written
        in the order they were added to the recipe.
        """
        if self.comment is not None:
            w(self.comment)
        for resource in self.resources:
            w(resource.dumps(inline))

    def dumps(self):
        """
        Generate a string containing Chef code and all file contents.
        """
        out = []
        return ''.join(out)

    def dumpf(self, gzip=False):
        """
        Generate files containing Chef code and templates.  The directory
        structure generated is that of a cookbook with a default recipe and
        cookbook files.
        """
        os.mkdir(self.name)
        f = codecs.open(os.path.join(self.name, 'metadata.rb'), 'w', encoding='utf-8')
        f.close()
        os.mkdir(os.path.join(self.name, 'recipes'))
        filename = os.path.join(self.name, 'recipes/default.rb')
        f = codecs.open(filename, 'w', encoding='utf-8')
        self._dump(f.write, inline=False)
        f.close()
        for resource in self.resources:
            if 'cookbook_file' != resource.type:
                continue
            pathname = os.path.join(self.name, 'files/default',
                resource.name[1:])
            try:
                os.makedirs(os.path.dirname(pathname))
            except OSError as e:
                if errno.EEXIST != e.errno:
                    raise e
            if isinstance(resource.content, unicode):
                f = codecs.open(pathname, 'w', encoding='utf-8')
            else:
                f = open(pathname, 'w')
            f.write(resource.content)
            f.close()
        if gzip:
            filename = 'chef-{0}.tar.gz'.format(self.name)
            tarball = tarfile.open(filename, 'w:gz')
            tarball.add(self.name)
            tarball.close()
            return filename
        return filename


class Resource(dict):
    """
    A Chef resource has a type, a name, and some parameters.  Nothing has
    to be unique as resources are dealt with in order rather than by building
    a dependency graph.

    """

    def __init__(self, type, name, **kwargs):
        """
        Don't instantiate this class directly.  Instead, use the methods made
        available in the Cookbook class.
        """
        super(Resource, self).__init__(**kwargs)
        self.type = type
        self.name = name

    @classmethod
    def _dumps(cls, value, recursive=False):
        """
        Return a value as it should be written.  If the value starts with
        a ':', it will be written as-is.  Otherwise, it will be written as
        a string.
        """
        if value is None:
            return 'nil'
        elif True == value:
            return 'true'
        elif False == value:
            return 'false'
        elif any([isinstance(value, t) for t in (int, long, float)]):
            return value
        elif 1 < len(value) and ':' == value[0]:
            return value
        elif hasattr(value, 'bare') or isinstance(value, util.BareString):
            return value
        elif isinstance(value, cls):
            return repr(value)
        elif isinstance(value, list) or isinstance(value, tuple):
            s = ', '.join([cls._dumps(v, True) for v in value])
            if recursive:
                return '[' + s + ']'
            else:
                return s
        return repr(unicode(value).replace(u'#{', u'\\#{'))[1:]

    def dumps(self, inline=False):
        """
        Stringify differently depending on the number of options so the
        output always looks like Ruby code should look.  Parentheses are
        always employed here due to grammatical inconsistencies when using
        braces surrounding a block.
        """
        if 0 == len(self):
            return u'{0}({1})\n'.format(self.type, self._dumps(self.name))
        elif 1 == len(self):
            key, value = self.items()[0]
            return u'{0}({1}) {{ {2} {3} }}\n'.format(self.type,
                                                      self._dumps(self.name),
                                                      key,
                                                      self._dumps(value))
        else:
            out = [u'{0}({1}) do\n'.format(self.type, self._dumps(self.name))]
            for key, value in sorted(self.iteritems()):
                out.append(u'  {0} {1}\n'.format(key, self._dumps(value)))
            out.append('end\n')
            return ''.join(out)


class File(Resource):
    """
    Special Chef file or cookbook_file resource.
    """

    def __init__(self, name, content=None, **kwargs):
        """
        File resources handle their content explicitly because in some
        cases it is not written as a normal parameter.
        """
        super(File, self).__init__('file', name, **kwargs)
        self.content = content

    def dumps(self, inline=False):
        """
        Decide whether to write as a file with content or a cookbook_file
        that leaves its content to be dealt with later.
        """
        if inline:
            if self.content is not None:
                self['content'] = self.content
                del self.content
            self.type = 'file'
            del self['source']
        elif self.content is not None and 'source' in self:
            self.type = 'cookbook_file'
        return super(File, self).dumps(inline)



from collections import defaultdict
import fnmatch
import glob
import json
import logging
import os
import os.path
import re
import subprocess

from blueprint import deps
from blueprint import util


# The default list of ignore patterns.  Typically, the value of each key
# will be False.  Providing True will negate the meaning of the pattern
# and cause matching files to be included in blueprints.
#
# XXX Update `blueprintignore`(5) if you make changes here.
IGNORE = {'*~': False,
          '*.blueprint-template.*': False,
          '*.dpkg-*': False,
          '/etc/.git': False,
          '/etc/.pwd.lock': False,
          '/etc/X11/default-display-manager': False,
          '/etc/adjtime': False,
          '/etc/alternatives': False,
          '/etc/apparmor': False,
          '/etc/apparmor.d': False,
          '/etc/blkid/blkid.tab': False,
          '/etc/ca-certificates.conf': False,
          '/etc/console-setup': False,

          # TODO Only if it's a symbolic link to ubuntu.
          '/etc/dpkg/origins/default': False,

          '/etc/fstab': False,
          '/etc/group-': False,
          '/etc/group': False,
          '/etc/gshadow-': False,
          '/etc/gshadow': False,
          '/etc/hostname': False,
          '/etc/init.d/.legacy-bootordering': False,
          '/etc/initramfs-tools/conf.d/resume': False,
          '/etc/ld.so.cache': False,
          '/etc/localtime': False,
          '/etc/lvm/cache': False,
          '/etc/mailcap': False,
          '/etc/mtab': False,
          '/etc/modules': False,

          # TODO Only if it's a symbolic link to /var/run/motd.
          '/etc/motd': False,

          '/etc/network/interfaces': False,
          '/etc/passwd-': False,
          '/etc/passwd': False,
          '/etc/pki/rpm-gpg': True,
          '/etc/popularity-contest.conf': False,
          '/etc/prelink.cache': False,
          '/etc/resolv.conf': False,  # Most people use the defaults.
          '/etc/rc.d': False,
          '/etc/rc0.d': False,
          '/etc/rc1.d': False,
          '/etc/rc2.d': False,
          '/etc/rc3.d': False,
          '/etc/rc4.d': False,
          '/etc/rc5.d': False,
          '/etc/rc6.d': False,
          '/etc/rcS.d': False,
          '/etc/shadow-': False,
          '/etc/shadow': False,
          '/etc/ssh/ssh_host_key*': False,
          '/etc/ssh/ssh_host_*_key*': False,
          '/etc/ssl/certs': False,
          '/etc/sysconfig/clock': False,
          '/etc/sysconfig/i18n': False,
          '/etc/sysconfig/keyboard': False,
          '/etc/sysconfig/network': False,
          '/etc/sysconfig/network-scripts': False,
          '/etc/timezone': False,
          '/etc/udev/rules.d/70-persistent-*.rules': False,
          '/etc/yum.repos.d': True}


CACHE = '/tmp/blueprintignore'


def defaults():
    """
    Parse `/etc/blueprintignore` and `~/.blueprintignore` to build the
    default `Rules` object.
    """
    r = None

    # Check for a fresh cache of the complete blueprintignore(5) rules.
    if _mtime('/etc/blueprintignore') < _mtime(CACHE) \
    and _mtime(os.path.expanduser('~/.blueprintignore')) < _mtime(CACHE) \
    and _mtime(__file__) < _mtime(CACHE):
        try:
            r = Rules(json.load(open(CACHE)))
            logging.info('using cached blueprintignore(5) rules')
            return r
        except (OSError, ValueError):
            pass

    # Cache things that are ignored by default first.
    r = Rules({
        'file': IGNORE.items(),
        'package': [('apt', package, False) for package in _apt()] +
                   [('yum', package, False) for package in _yum()],
        'service': [('sysvinit', 'skeleton', False)],
        'source': [('/', False),
                   ('/usr/local', True)],
    })

    # Cache the patterns stored in the blueprintignore files.
    logging.info('parsing blueprintignore(5) rules')
    try:
        for pathname in ['/etc/blueprintignore',
                         os.path.expanduser('~/.blueprintignore')]:
            r.parse(open(pathname), negate=True)

    except IOError:
        pass

    # Store the cache to disk.
    f = _cache_open(CACHE, 'w')
    json.dump(r, f, indent=2, sort_keys=True)
    f.close()

    return r


def none():
    """
    Build a `Rules` object that ignores every resource.
    """
    return Rules({'file': [('*', False)],
                  'package': [('*', '*', False)],
                  'service': [('*', '*', False)],
                  'source': [('/', False)]})


def _apt():
    """
    Return the set of packages that should never appear in a blueprint because
    they're already guaranteed (to some degree) to be there.
    """

    CACHE = '/tmp/blueprint-apt-exclusions'

    # Read from a cached copy.
    try:
        return set([line.rstrip() for line in open(CACHE)])
    except IOError:
        pass
    logging.info('searching for APT packages to exclude')

    # Start with the root packages for the various Ubuntu installations.
    s = set(['grub-pc',
             'installation-report',
             'language-pack-en',
             'language-pack-gnome-en',
             'linux-generic-pae',
             'linux-server',
             'os-prober',
             'ubuntu-desktop',
             'ubuntu-minimal',
             'ubuntu-standard',
             'wireless-crda'])

    # Find the essential and required packages.  Every server's got 'em, no
    # one wants to muddle their blueprint with 'em.
    for field in ('Essential', 'Priority'):
        try:
            p = subprocess.Popen(['dpkg-query',
                                  '-f=${{Package}} ${{{0}}}\n'.format(field),
                                  '-W'],
                                 close_fds=True,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
        except OSError:
            _cache_open(CACHE, 'w').close()
            return s
        for line in p.stdout:
            try:
                package, property = line.rstrip().split()
                if property in ('yes', 'important', 'required', 'standard'):
                    s.add(package)
            except ValueError:
                pass

    # Walk the dependency tree all the way to the leaves.
    s = deps.apt(s)

    # Write to a cache.
    logging.info('caching excluded APT packages')
    f = _cache_open(CACHE, 'w')
    for package in sorted(s):
        f.write('{0}\n'.format(package))
    f.close()

    return s


def _yum():
    """
    Return the set of packages that should never appear in a blueprint because
    they're already guaranteed (to some degree) to be there.
    """

    CACHE = '/tmp/blueprint-yum-exclusions'

    # Read from a cached copy.
    try:
        return set([line.rstrip() for line in open(CACHE)])
    except IOError:
        pass
    logging.info('searching for Yum packages to exclude')

    # Start with a few groups that install common packages.
    s = set(['gpg-pubkey'])
    pattern = re.compile(r'^   (\S+)')
    try:
        p = subprocess.Popen(['yum', 'groupinfo',
                              'core','base', 'gnome-desktop'],
                             close_fds=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
    except OSError:
        _cache_open(CACHE, 'w').close()
        return s
    for line in p.stdout:
        match = pattern.match(line)
        if match is not None:
            s.add(match.group(1))

    # Walk the dependency tree all the way to the leaves.
    s = deps.yum(s)

    # Write to a cache.
    logging.info('caching excluded Yum packages')
    f = _cache_open(CACHE, 'w')
    for package in sorted(s):
        f.write('{0}\n'.format(package))
    f.close()

    return s


def _cache_open(pathname, mode):
    f = open(pathname, mode)
    if util.via_sudo():
        uid = int(os.environ['SUDO_UID'])
        gid = int(os.environ['SUDO_GID'])
        os.fchown(f.fileno(), uid, gid)
    return f


def _mtime(pathname):
    try:
        return os.stat(pathname).st_mtime
    except OSError:
        return 0


class Rules(defaultdict):
    """
    Ordered lists of rules for ignoring/unignoring particular resources.
    This is used for both `blueprintignore`(5) and `blueprint-rules`(1).
    """

    def __init__(self, *args, **kwargs):
        super(Rules, self).__init__(list, *args, **kwargs)

    def _ignore_pathname(self, restype, dirname, pathname, ignored=False):
        """
        Return `True` if the `gitignore`(5)-style `~/.blueprintignore`
        file says the given file should be ignored.  The starting state
        of the file may be overridden by setting `ignored` to `True`.
        """
        pathname = util.unicodeme(pathname)

        # Determine if the `pathname` matches the `pattern`.  `filename` is
        # given as a convenience.  See `gitignore`(5) for the rules in play.
        def match(filename, pathname, pattern):
            dir_only = '/' == pattern[-1]
            pattern = pattern.rstrip('/')
            if '/' not in pattern:
                if fnmatch.fnmatch(filename, pattern):
                    return os.path.isdir(pathname) if dir_only else True
            else:
                for p in glob.glob(os.path.join(dirname, pattern)):
                    p = util.unicodeme(p)
                    if pathname == p or pathname.startswith('{0}/'.format(p)):
                        return os.path.isdir(pathname) if dir_only else True
            return False

        # Iterate over exclusion rules until a match is found.  Then iterate
        # over inclusion rules that appear later.  If there are no matches,
        # include the file.  If only an exclusion rule matches, exclude the
        # file.  If an inclusion rule also matches, include the file.
        filename = os.path.basename(pathname)
        for pattern, negate in self[restype]:
            if ignored != negate or not match(filename, pathname, pattern):
                continue
            ignored = not ignored

        return ignored

    def ignore_file(self, pathname, ignored=False):
        """
        Return `True` if the given pathname should be ignored.
        """
        return self._ignore_pathname('file', '/etc', pathname, ignored)

    def ignore_package(self, manager, package, ignored=False):
        """
        Iterate over package exclusion rules looking for exact matches. As
        with files, search for a negated rule after finding a match. Return
        `True` to indicate the package should be ignored.
        """
        for m, p, negate in self['package']:
            if ignored != negate \
            or manager != m and '*' != m \
            or package != p and '*' != p:
                continue
            ignored = not ignored
        return ignored

    def ignore_service(self, manager, service, ignored=False):
        """
        Return `True` if a given service should be ignored.
        """
        for m, s, negate in self['service']:
            if ignored != negate \
            or manager != m and '*' != m \
            or service != s and '*' != s:
                continue
            ignored = not ignored
        return ignored

    def ignore_source(self, pathname, ignored=False):
        """
        Return `True` if the given pathname should be ignored.  Negated rules
        on directories will create new source tarballs.  Other rules will
        ignore files within those tarballs.
        """
        return self._ignore_pathname('source', '/', pathname, ignored)


    def parse(self, f, negate=False):
        """
        Parse rules from the given file-like object.  This is used both for
        `blueprintignore`(5) and for `blueprint-rules`(1).
        """
        for pattern in f:
            pattern = pattern.rstrip()

            # Comments and blank lines.
            if '' == pattern or '#' == pattern[0]:
                continue

            # Negated lines.
            if '!' == pattern[0]:
                pattern = pattern[1:]
                ignored = negate
            else:
                ignored = not negate

            # Normalize file resources, which don't need the : and type
            # qualifier, into the same format as others, like packages.
            if ':' == pattern[0]:
                try:
                    restype, pattern = pattern[1:].split(':', 2)
                except ValueError:
                    continue
            else:
                restype = 'file'

            # Ignore a package and its dependencies or unignore a single
            # package.  Empirically, the best balance of power and
            # granularity comes from this arrangement.  Take
            # build-esseantial's mutual dependence with dpkg-dev as an
            # example of why.
            if 'package' == restype:
                try:
                    manager, package = pattern.split('/')
                except ValueError:
                    logging.warning('invalid package rule "{0}"'.
                                    format(pattern))
                    continue
                self['package'].append((manager, package, ignored))
                if not ignored:
                    for dep in getattr(deps,
                                       manager,
                                       lambda(arg): [])(package):
                        self['package'].append((manager, dep, ignored))

            elif 'service' == restype:
                try:
                    manager, service = pattern.split('/')
                except ValueError:
                    logging.warning('invalid service rule "{0}"'.
                                    format(pattern))
                    continue
                self['service'].append((manager, service, ignored))

            # Ignore or unignore a file, glob, or directory tree.
            else:
                self[restype].append((pattern, ignored))

        return self



"""
Search for Python packages to include in the blueprint.
"""

import glob
import logging
import os
import re
import subprocess


# Precompile a pattern to extract the manager from a pathname.
pattern_manager = re.compile(r'lib/(python[^/]*)/(dist|site)-packages')

# Precompile patterns for differentiating between packages built by
# `easy_install` and packages built by `pip`.
pattern_egg = re.compile(r'\.egg$')
pattern_egginfo = re.compile(r'\.egg-info$')

# Precompile a pattern for extracting package names and version numbers.
pattern = re.compile(r'^([^-]+)-([^-]+).*\.egg(-info)?$')


def pypi(b, r):
    logging.info('searching for Python packages')

    # Look for packages in the typical places.  `pip` has its `freeze`
    # subcommand but there is no way but diving into the directory tree to
    # figure out what packages were `easy_install`ed.  If `VIRTUAL_ENV`
    # appears in the environment, treat the directory it names just like
    # the global package directories.
    globnames = ['/usr/lib/python*/dist-packages',
                 '/usr/lib/python*/site-packages',
                 '/usr/local/lib/python*/dist-packages',
                 '/usr/local/lib/python*/site-packages']
    virtualenv = os.getenv('VIRTUAL_ENV')
    if virtualenv is not None:
        globnames.extend(['{0}/lib/python*/dist-packages'.format(virtualenv),
                          '{0}/lib/python*/dist-packages'.format(virtualenv)])
    for globname in globnames:
        for dirname in glob.glob(globname):
            manager = pattern_manager.search(dirname).group(1)
            for entry in os.listdir(dirname):
                match = pattern.match(entry)
                if match is None:
                    continue
                package, version = match.group(1, 2)
                pathname = os.path.join(dirname, entry)

                # Symbolic links indicate this is actually a system package
                # that injects files into the PYTHONPATH.
                if os.path.islink(pathname):
                    continue

                # Assume this is a Debian-based system and let `OSError`
                # looking for `dpkg-query` prove this is RPM-based.  In
                # that case, the dependencies get a bit simpler.
                try:
                    _dpkg_query(b, r,
                                manager, package, version,
                                entry, pathname)
                except OSError:
                    try:
                        _rpm(b, r, manager, package, version, entry, pathname)
                    except OSError:
                        logging.warning('neither dpkg nor rpm found')


def _dpkg_query(b, r, manager, package, version, entry, pathname):
    """
    Resolve dependencies on Debian-based systems.
    """

    # If this Python package is actually part of a system
    # package, abandon it.
    p = subprocess.Popen(['dpkg-query', '-S', pathname],
                         close_fds=True,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    p.communicate()
    if 0 == p.returncode:
        return

    # This package was installed via `easy_install`.  Make
    # sure its version of Python is in the blueprint so it
    # can be used as a package manager.
    if pattern_egg.search(entry):
        p = subprocess.Popen(['dpkg-query', '-f=${Version}', '-W', manager],
                             close_fds=True,
                             stdout=subprocess.PIPE)
        stdout, stderr = p.communicate()
        if 0 != p.returncode:
            return
        versions = b.packages['apt'][manager]
        if stdout not in versions:
            versions.add(stdout)
        if not r.ignore_package(manager, package):
            b.add_package(manager, package, version)

    # This package was installed via `pip`.  Figure out how
    # `pip` was installed and use that as this package's
    # manager.
    elif pattern_egginfo.search(entry) and os.path.exists(
        os.path.join(pathname, 'installed-files.txt')):
        p = subprocess.Popen(['dpkg-query', '-W', 'python-pip'],
                             close_fds=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        p.communicate()
        if 0 != p.returncode:
            if not r.ignore_package('pip', package):
                b.add_package('pip', package, version)
        else:
            if not r.ignore_package('python-pip', package):
                b.add_package('python-pip', package, version)


def _rpm(b, r, manager, package, version, entry, pathname):
    """
    Resolve dependencies on RPM-based systems.
    """

    # If this Python package is actually part of a system
    # package, abandon it.
    p = subprocess.Popen(['rpm', '-qf', pathname],
                         close_fds=True,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    p.communicate()
    if 0 == p.returncode:
        return

    # This package was installed via `easy_install`.  Make
    # sure Python is in the blueprint so it can be used as
    # a package manager.
    if pattern_egg.search(entry):
        p = subprocess.Popen(['rpm',
                              '--qf=%{VERSION}-%{RELEASE}.%{ARCH}',
                              '-q',
                              'python'],
                             close_fds=True,
                             stdout=subprocess.PIPE)
        stdout, stderr = p.communicate()
        if 0 != p.returncode:
            return
        versions = b.packages['yum']['python']
        if stdout not in versions:
            versions.add(stdout)
        if not r.ignore_package('python', package):
            b.add_package('python', package, version)

    # This package was installed via `pip`.  Figure out how
    # `pip` was installed and use that as this package's
    # manager.
    elif pattern_egginfo.search(entry) and os.path.exists(
        os.path.join(pathname, 'installed-files.txt')):
        p = subprocess.Popen(['rpm', '-q', 'python-pip'],
                             close_fds=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        p.communicate()
        if 0 != p.returncode:
            if not r.ignore_package('pip', package):
                b.add_package('pip', package, version)
        else:
            if not r.ignore_package('python-pip', package):
                b.add_package('python-pip', package, version)



import logging
import re
import subprocess


def apt(s):
    """
    Walk the dependency tree of all the packages in set s all the way to
    the leaves.  Return the set of s plus all their dependencies.
    """
    logging.debug('searching for APT dependencies')
    if not isinstance(s, set):
        s = set([s])
    tmp_s = s
    pattern_sub = re.compile(r'\([^)]+\)')
    pattern_split = re.compile(r'[,\|]')
    while 1:
        new_s = set()
        for package in tmp_s:
            p = subprocess.Popen(
                ['dpkg-query',
                 '-f', '${Pre-Depends}\n${Depends}\n${Recommends}\n',
                 '-W', package],
                close_fds=True, stdout=subprocess.PIPE)
            for line in p.stdout:
                line = line.strip()
                if '' == line:
                    continue
                for part in pattern_split.split(pattern_sub.sub('', line)):
                    new_s.add(part.strip())

        # If there is to be a next iteration, `new_s` must contain some
        # packages not yet in `s`.
        tmp_s = new_s - s
        if 0 == len(tmp_s):
            break
        s |= new_s

    return s


def yum(s):
    """
    Walk the dependency tree of all the packages in set s all the way to
    the leaves.  Return the set of s plus all their dependencies.
    """
    logging.debug('searching for Yum dependencies')

    if not hasattr(yum, '_cache'):
        yum._cache = {}
        try:
            p = subprocess.Popen(['rpm',
                                  '-qa',
                                  '--qf=%{NAME}\x1E[%{PROVIDES}\x1F]\n'],
                                 close_fds=True,
                                 stdout=subprocess.PIPE)
            for line in p.stdout:
                name, caps = line.rstrip().split('\x1E')
                yum._cache.update([(cap, name) for cap in caps.split('\x1F')])
        except OSError:
            pass

    if not isinstance(s, set):
        s = set([s])

    tmp_s = s
    while 1:
        new_s = set()
        for package in tmp_s:
            try:
                p = subprocess.Popen(['rpm', '-qR', package],
                                     close_fds=True,
                                     stdout=subprocess.PIPE)
            except OSError:
                continue
            for line in p.stdout:
                cap = line.rstrip()[0:line.find(' ')]
                if 'rpmlib' == cap[0:6]:
                    continue
                try:
                    new_s.add(yum._cache[cap])
                except KeyError:
                    try:
                        p2 = subprocess.Popen(['rpm',
                                               '-q',
                                               '--qf=%{NAME}',
                                               '--whatprovides',
                                               cap],
                                              close_fds=True,
                                              stdout=subprocess.PIPE)
                        stdout, stderr = p2.communicate()
                        yum._cache[cap] = stdout
                        new_s.add(stdout)
                    except OSError:
                        pass

        # If there is to be a next iteration, `new_s` must contain some
        # packages not yet in `s`.
        tmp_s = new_s - s
        if 0 == len(tmp_s):
            break
        s |= new_s

    return s



import base64
from flask import Flask, Response, abort, redirect, render_template, request
import hashlib
import json
import os
import re
import sys
import urlparse

from blueprint import Blueprint
from blueprint import cfg
import backend
import librato
import statsd


app = Flask(__name__)


def _blueprint(secret, name):
    """
    Fetch a blueprint from S3 and turn it into a real Blueprint object.

    The name can't be given as a kwarg or Blueprint.__init__ will go
    looking for the JSON in Git.
    """
    data = backend.get_blueprint(secret, name)
    if data is None:
        return None
    elif data is False:
        return False
    b = Blueprint()
    b.name = name
    b.update(json.loads(data))
    return b


def _server():
    """
    Reconstitute the name of this Blueprint I/O Server.
    """
    return urlparse.urlunparse((request.environ.get('wsgi.url_scheme',
                                                    'https'),
                                request.environ.get('HTTP_HOST',
                                                    'devstructure.com'),
                                '', '', '', ''))


class MeteredResponse(Response):
    """
    HTTP responses that account for the outbound bandwidth they consume.
    """
    def __init__(self, *args, **kwargs):
        super(MeteredResponse, self).__init__(*args, **kwargs)
        content_length = len(kwargs.get('response', ''))
        if 0 < content_length:
            librato.count('blueprint-io-server.bandwidth.out', content_length)
            statsd.update('blueprint-io-server.bandwidth.out', content_length)


@app.errorhandler(400)
def bad_request(e):
    return MeteredResponse(response='',
                           status=400,
                           content_type='text/plain')


@app.errorhandler(404)
def not_found(e):
    return MeteredResponse(response='',
                           status=404,
                           content_type='text/plain')


@app.errorhandler(502)
def bad_gateway(e):
    return MeteredResponse(response='',
                           status=502,
                           content_type='text/plain')


def validate_secret(secret):
    if re.match(r'^[0-9A-Za-z_-]{64}$', secret) is None:
        abort(400)


def validate_name(name):
    if re.search(r'[/ \t\r\n]', name) is not None:
        abort(400)


def validate_sha(sha):
    if re.match(r'^[0-9a-f]{40}$', sha) is None:
        abort(400)


def validate_content_length():
    if cfg.getint('io', 'max_content_length') < request.content_length:
        abort(413)


@app.route('/secret', methods=['GET'])
def secret():
    while 1:
        s = base64.urlsafe_b64encode(os.urandom(48))
        try:
            iter(backend.list(s)).next()
        except StopIteration:
            break
    return MeteredResponse(response='{0}\n'.format(s),
                           status=201,
                           content_type='text/plain')


browser_pattern = re.compile(r'Chrome|Gecko|Microsoft|Mozilla|Safari|WebKit')


@app.route('/<secret>/<name>', methods=['GET'])
def get_blueprint(secret, name):
    validate_secret(secret)
    validate_name(name)

    content_length = backend.head_blueprint(secret, name)
    if content_length is None:
        abort(404)

    # Pretty HTML for browsers.
    if browser_pattern.search(request.environ.get('HTTP_USER_AGENT', '')) \
       or 'html' == request.args.get('format'):
        librato.count('blueprint-io-server.renders')
        statsd.increment('blueprint-io-server.renders')
        return render_template('blueprint.html', b=_blueprint(secret, name))

    # Raw JSON for everybody else.
    else:
        librato.count('blueprint-io-server.requests.get')
        statsd.increment('blueprint-io-server.requests.get')
        librato.count('blueprint-io-server.bandwidth.out', content_length)
        statsd.update('blueprint-io-server.bandwidth.out', content_length)
        return redirect(backend.url_for_blueprint(secret, name), code=301)


@app.route('/<secret>/<name>', methods=['PUT'])
def put_blueprint(secret, name):
    validate_secret(secret)
    validate_name(name)

    librato.count('blueprint-io-server.bandwidth.in', request.content_length)
    statsd.update('blueprint-io-server.bandwidth.in', request.content_length)
    validate_content_length()

    # Validate the blueprint JSON format.  This could stand more rigor
    # or, dare I say it, a schema?
    try:
        for k in request.json.iterkeys():
            if k not in ('arch', 'files', 'packages', 'services', 'sources'):
                abort(400)
    except ValueError:
        abort(400)

    # Remove tarballs referenced by the old blueprint but not the new one.
    b = _blueprint(secret, name)
    if b is not None and b is not False:
        for filename in set(b.sources.itervalues()) - \
                        set(request.json.get('sources', {}).itervalues()):
            backend.delete_tarball(secret, name, filename[0:-4])

    # Store the blueprint JSON in S3.
    if not backend.put_blueprint(secret, name, request.data):
        abort(502)

    return MeteredResponse(response='',
                           status=202,
                           content_type='text/plain')


@app.route('/<secret>/<name>/<sha>.tar', methods=['GET'])
def get_tarball(secret, name, sha):
    validate_secret(secret)
    validate_name(name)
    sha = sha.lower()
    validate_sha(sha)

    content_length = backend.head_tarball(secret, name, sha)
    if content_length is None:
        abort(404)

    librato.count('blueprint-io-server.requests.get')
    statsd.increment('blueprint-io-server.requests.get')
    librato.count('blueprint-io-server.bandwidth.out', content_length)
    statsd.update('blueprint-io-server.bandwidth.out', content_length)

    return redirect(backend.url_for_tarball(secret, name, sha), code=301)


@app.route('/<secret>/<name>/<sha>.tar', methods=['PUT'])
def put_tarball(secret, name, sha):
    validate_secret(secret)
    validate_name(name)
    sha = sha.lower()
    validate_sha(sha)

    librato.count('blueprint-io-server.bandwidth.in', request.content_length)
    statsd.update('blueprint-io-server.bandwidth.in', request.content_length)
    validate_content_length()

    # Validate the tarball content.
    if hashlib.sha1(request.data).hexdigest() != sha:
        abort(400)

    # Ensure the tarball appears in the blueprint.
    b = _blueprint(secret, name)
    if b is None:
        abort(404)
    elif b is False:
        abort(502)
    if '{0}.tar'.format(sha) not in b.sources.itervalues():
        abort(400)

    # Store the tarball in S3.
    if not backend.put_tarball(secret, name, sha, request.data):
        abort(502)

    return MeteredResponse(response='',
                           status=202,
                           content_type='text/plain')


@app.route('/<secret>/<name>/<name2>.sh', methods=['GET'])
def sh(secret, name, name2):
    if 'user-data' == name2:
        return user_data(secret, name)
    if name != name2:
        abort(400)
    validate_secret(secret)
    validate_name(name)

    # Generate POSIX shell code from the blueprint.
    b = _blueprint(secret, name)
    if b is None:
        abort(404)
    elif b is False:
        abort(502)
    s = b.sh(server=_server(), secret=secret)
    s.out.insert(0, '#!/bin/sh\n\n')
    return MeteredResponse(response=s.dumps(),
                           status=200,
                           content_type='text/plain')


@app.route('/<secret>/<name>/user-data.sh', methods=['GET'])
def user_data(secret, name):
    validate_secret(secret)
    validate_name(name)
    b = _blueprint(secret, name)
    if b is None:
        abort(404)
    elif b is False:
        abort(502)
    return MeteredResponse(response="""#!/bin/sh

set -e

TMPDIR="$(mktemp -d)"
cd "$TMPDIR"
trap "rm -rf \\"$TMPDIR\\"" EXIT

wget "{0}/{1}/{2}/{2}.sh"

sh "$(ls)"
""".format(_server(), secret, name),
                           status=200,
                           content_type='text/plain')


if '__main__' == __name__:
    app.run(host='0.0.0.0', debug=True)



from ConfigParser import ConfigParser
from collections import defaultdict
import copy
import json
import logging
import os.path
import re
import sys

# This must be called early - before the rest of the blueprint library loads.
logging.basicConfig(format='# [blueprint] %(message)s',
                    level=logging.INFO)

import git
import rules
import util
import walk


DEFAULTS = {'io': {'max_content_length': 67108864,
                   'server': 'https://devstructure.com'},
            's3': {'region': 'US',
                   'use_https': True},
            'statsd': {'port': 8125}}


cfg = ConfigParser()
for section, options in DEFAULTS.iteritems():
    cfg.add_section(section)
    for option, value in options.iteritems():
        cfg.set(section, option, str(value))
cfg.read(['/etc/blueprint.cfg',
          os.path.expanduser('~/.blueprint.cfg')])


class NameError(ValueError):
    pass


class NotFoundError(KeyError):
    pass


class Blueprint(dict):

    DISCLAIMER = """#
# Automatically generated by blueprint(7).  Edit at your own risk.
#
"""

    @classmethod
    def checkout(cls, name, commit=None):
        git.init()
        if commit is None:
            commit = git.rev_parse('refs/heads/{0}'.format(name))
            if commit is None:
                raise NotFoundError(name)
        tree = git.tree(commit)
        blob = git.blob(tree, 'blueprint.json')
        content = git.content(blob)
        return cls(name, commit, **json.loads(content))

    @classmethod
    def create(cls, name):
        b = cls(name)
        r = rules.defaults()
        import backend
        for funcname in backend.__all__:
            getattr(backend, funcname)(b, r)
        import services
        services.services(b)
        return b

    @classmethod
    def destroy(cls, name):
        """
        Destroy the named blueprint.
        """
        if not os.path.isdir(git.repo()):
            raise NotFoundError(name)
        try:
            git.git('branch', '-D', name)
        except:
            raise NotFoundError(name)

    @classmethod
    def iter(cls):
        """
        Yield the name of each blueprint.
        """
        if not os.path.isdir(git.repo()):
            return
        status, stdout = git.git('branch')
        for line in stdout.splitlines():
            yield line.strip()

    @classmethod
    def load(cls, f, name=None):
        """
        Instantiate and return a Blueprint object from a file-like object
        from which valid blueprint JSON may be read.
        """
        return cls(name, **json.load(f))

    @classmethod
    def loads(cls, s, name=None):
        """
        Instantiate and return a Blueprint object from a string containing
        valid blueprint JSON.
        """
        return cls(name, **json.loads(s))

    @classmethod
    def rules(cls, r, name=None):
        b = cls(name)
        import backend
        for funcname in backend.__all__:
            getattr(backend, funcname)(b, r)
        import services
        services.services(b)
        return b

    def __init__(self, name=None, commit=None, *args, **kwargs):
        """
        Construct a blueprint.  Extra arguments are used to create a `dict`
        which is then sent through the `blueprint`(5) algorithm to be injested
        into this `Blueprint` object with the proper types.  (The structure
        makes heavy use of `defaultdict` and `set`).
        """
        self.name = name
        self._commit = commit
        def file(pathname, f):
            self.add_file(pathname, **f)
        def package(manager, package, version):
            self.add_package(manager, package, version)
        def service(manager, service):
            self.add_service(manager, service)
        def service_file(manager, service, pathname):
            self.add_service_file(manager, service, pathname)
        def service_package(manager, service, package_manager, package):
            self.add_service_package(manager,
                                     service,
                                     package_manager,
                                     package)
        def service_source(manager, service, dirname):
            self.add_service_source(manager, service, dirname)
        def source(dirname, filename, gen_content, url):
            if url is not None:
                self.add_source(dirname, url)
            elif gen_content is not None:
                self.add_source(dirname, filename)
        walk.walk(dict(*args, **kwargs),
                  file=file,
                  package=package,
                  service=service,
                  service_file=service_file,
                  service_package=service_package,
                  service_source=service_source,
                  source=source)

    def __sub__(self, other):
        """
        Subtracting one blueprint from another allows blueprints to remain
        free of superfluous packages from the base installation.  It takes
        three passes through the package tree.  The first two remove
        superfluous packages and the final one accounts for some special
        dependencies by adding them back to the tree.
        """
        b = copy.deepcopy(self)

        # Compare file contents and metadata.  Keep files that differ.
        for pathname, file in self.files.iteritems():
            if other.files.get(pathname, {}) == file:
                del b.files[pathname]

        # The first pass removes all duplicate packages that are not
        # themselves managers.  Allowing multiple versions of the same
        # packages complicates things slightly.  For each package, each
        # version that appears in the other blueprint is removed from
        # this blueprint.  After that is finished, this blueprint is
        # normalized.  If no versions remain, the package is removed.
        def package(manager, package, version):
            if package in b.packages:
                return
            try:
                b_packages = b.packages[manager]
            except KeyError:
                return
            if manager in b_packages:
                return
            if package not in b_packages:
                return
            try:
                b_versions = b_packages[package]
            except KeyError:
                return
            b_versions.discard(version)
            if 0 == len(b_versions):
                del b_packages[package]
            else:
                b_packages[package] = b_versions
        other.walk(package=package)

        # The second pass removes managers that manage no packages, a
        # potential side-effect of the first pass.  This step must be
        # applied repeatedly until the blueprint reaches a steady state.
        def package(manager, package, version):
            if package not in b.packages:
                return
            if 0 == len(b.packages[package]):
                del b.packages[package]
                del b.packages[self.managers[package]][package]
        while 1:
            l = len(b.packages)
            other.walk(package=package)
            if len(b.packages) == l:
                break

        # The third pass adds back special dependencies like `ruby*-dev`.
        # It isn't apparent from the rules above that a manager like RubyGems
        # needs more than just itself to function.  In some sense, this might
        # be considered a missing dependency in the Debian archive but in
        # reality it's only _likely_ that you need `ruby*-dev` to use
        # `rubygems*`.
        def after_packages(manager):
            if manager not in b.packages:
                return

            deps = {r'^python(\d+(?:\.\d+)?)$': ['python{0}',
                                                 'python{0}-dev',
                                                 'python',
                                                 'python-devel'],
                    r'^ruby(\d+\.\d+(?:\.\d+)?)$': ['ruby{0}-dev'],
                    r'^rubygems(\d+\.\d+(?:\.\d+)?)$': ['ruby{0}',
                                                        'ruby{0}-dev',
                                                        'ruby',
                                                        'ruby-devel']}

            for pattern, packages in deps.iteritems():
                match = re.search(pattern, manager)
                if match is None:
                    continue
                for package in packages:
                    package = package.format(match.group(1))
                    for managername in ('apt', 'yum'):
                        mine = self.packages.get(managername, {}).get(package,
                                                                      None)
                        if mine is not None:
                            b.packages[managername][package] = mine
        other.walk(after_packages=after_packages)

        # Compare service metadata.  Keep services that differ.
        for manager, services in self.services.iteritems():
            for service, deps in services.iteritems():
                if other.services.get(manager, {}).get(service, {}) == deps:
                    del b.services[manager][service]
            if 0 == len(b.services[manager]):
                del b.services[manager]

        # Compare source tarball filenames, which indicate their content.
        # Keep source tarballs that differ.
        for dirname, filename in self.sources.iteritems():
            if other.sources.get(dirname, '') == filename:
                del b.sources[dirname]

        return b

    def get_name(self):
        return self._name
    def set_name(self, name):
        """
        Validate and set the blueprint name.
        """
        if name is not None and re.search(r'^$|^-$|[/ \t\r\n]', name):
            raise NameError('invalid blueprint name')
        self._name = name
    name = property(get_name, set_name)

    def get_arch(self):
        if 'arch' not in self:
            self['arch'] = None
        return self['arch']
    def set_arch(self, arch):
        self['arch'] = arch
    arch = property(get_arch, set_arch)

    @property
    def files(self):
        if 'files' not in self:
            self['files'] = defaultdict(dict)
        return self['files']

    @property
    def managers(self):
        """
        Build a hierarchy of managers for easy access when declaring
        dependencies.
        """
        if hasattr(self, '_managers'):
            return self._managers
        self._managers = {'apt': None, 'yum': None}

        def package(manager, package, version):
            if package in self.packages and manager != package:
                self._managers[package] = manager

        self.walk(package=package)
        return self._managers

    @property
    def packages(self):
        if 'packages' not in self:
            self['packages'] = defaultdict(lambda: defaultdict(set))
        return self['packages']

    @property
    def services(self):
        if 'services' not in self:
            self['services'] = defaultdict(lambda: defaultdict(dict))
        return self['services']

    @property
    def sources(self):
        if 'sources' not in self:
            self['sources'] = defaultdict(dict)
        return self['sources']

    def add_file(self, pathname, **kwargs):
        """
        Create a file resource.
        """
        self.files[pathname] = kwargs

    def add_package(self, manager, package, version):
        """
        Create a package resource.
        """
        self.packages[manager][package].add(version)

    def add_service(self, manager, service):
        """
        Create a service resource which depends on given files and packages.
        """

        # AWS cfn-init respects the enable and ensure parameters like Puppet
        # does.  Blueprint provides these parameters for interoperability.
        self.services[manager].setdefault(service, {'enable': True,
                                                    'ensureRunning': True})

    def add_service_file(self, manager, service, *args):
        """
        Add file dependencies to a service resource.
        """
        if 0 == len(args):
            return
        s = self.services[manager][service].setdefault('files', set())
        for dirname in args:
            s.add(dirname)

    def add_service_package(self, manager, service, package_manager, *args):
        """
        Add package dependencies to a service resource.
        """
        if 0 == len(args):
            return
        d = self.services[manager][service].setdefault('packages',
                                                       defaultdict(set))
        for package in args:
            d[package_manager].add(package)

    def add_service_source(self, manager, service, *args):
        """
        Add source tarball dependencies to a service resource.
        """
        if 0 == len(args):
            return
        s = self.services[manager][service].setdefault('sources', set())
        for dirname in args:
            s.add(dirname)

    def add_source(self, dirname, filename):
        """
        Create a source tarball resource.
        """
        self.sources[dirname] = filename

    def commit(self, message=''):
        """
        Create a new revision of this blueprint in the local Git repository.
        Include the blueprint JSON and any source archives referenced by
        the JSON.
        """
        git.init()
        refname = 'refs/heads/{0}'.format(self.name)
        parent = git.rev_parse(refname)

        # Start with an empty index every time.  Specifically, clear out
        # source tarballs from the parent commit.
        if parent is not None:
            for mode, type, sha, pathname in git.ls_tree(git.tree(parent)):
                git.git('update-index', '--force-remove', pathname)

        # Add `blueprint.json` to the index.
        f = open('blueprint.json', 'w')
        f.write(self.dumps())
        f.close()
        git.git('update-index', '--add', os.path.abspath('blueprint.json'))

        # Add source tarballs to the index.
        for filename in self.sources.itervalues():
            git.git('update-index', '--add', os.path.abspath(filename))

        # Add `/etc/blueprintignore` and `~/.blueprintignore` to the index.
        # Since adding extra syntax to this file, it no longer makes sense
        # to store it as `.gitignore`.
        f = open('blueprintignore', 'w')
        for pathname in ('/etc/blueprintignore',
                         os.path.expanduser('~/.blueprintignore')):
            try:
                f.write(open(pathname).read())
            except IOError:
                pass
        f.close()
        git.git('update-index', '--add', os.path.abspath('blueprintignore'))

        # Write the index to Git's object store.
        tree = git.write_tree()

        # Write the commit and update the tip of the branch.
        self._commit = git.commit_tree(tree, message, parent)
        git.git('update-ref', refname, self._commit)

    def normalize(self):
        """
        Remove superfluous empty keys to reduce variance in serialized JSON.
        """
        if 'arch' in self and self['arch'] is None:
            del self['arch']
        for key in ['files', 'packages', 'sources']:
            if key in self and 0 == len(self[key]):
                del self[key]

    def dumps(self):
        """
        Return a JSON serialization of this blueprint.  Make a best effort
        to prevent variance from run-to-run.
        """
        self.normalize()
        return util.json_dumps(self)

    def puppet(self, relaxed=False):
        """
        Generate Puppet code.
        """
        import frontend.puppet
        return frontend.puppet.puppet(self, relaxed)

    def chef(self, relaxed=False):
        """
        Generate Chef code.
        """
        import frontend.chef
        return frontend.chef.chef(self, relaxed)

    def sh(self,
           relaxed=False,
           server='https://devstructure.com',
           secret=None):
        """
        Generate shell code.
        """
        import frontend.sh
        return frontend.sh.sh(self, relaxed, server, secret)

    def blueprint_rules(self, relaxed=False):
        """
        Generate Blueprint rules.
        """
        import frontend.rules
        return frontend.rules.rules(self, relaxed)

    def cfn(self, relaxed=False):
        """
        Generate an AWS CloudFormation template.
        """
        import frontend.cfn
        return frontend.cfn.cfn(self, relaxed)

    def blueprintignore(self):
        """
        Return an open file pointer to the blueprint's blueprintignore file,
        which is suitable for passing back to `blueprint.rules.Rules.parse`.
        Prior to v3.0.9 this file was stored as .blueprintignore in the
        repository.  Prior to v3.0.4 this file was stored as .gitignore in
        the repository.
        """
        tree = git.tree(self._commit)
        blob = git.blob(tree, 'blueprintignore')
        if blob is None:
            blob = git.blob(tree, '.blueprintignore')
        if blob is None:
            blob = git.blob(tree, '.gitignore')
        if blob is None:
            return []
        return git.cat_file(blob)

    def walk(self, **kwargs):
        walk.walk(self, **kwargs)



import logging
import os
import os.path
import subprocess
import sys

from blueprint import util


class GitError(EnvironmentError):
    pass


def unroot():
    """
    Drop privileges gained through sudo(1).
    """
    if util.via_sudo():
        uid = int(os.environ['SUDO_UID'])
        gid = int(os.environ['SUDO_GID'])
        os.setgid(gid)
        os.setegid(gid)
        os.setuid(uid)
        os.seteuid(uid)


def init():
    """
    Initialize the Git repository.
    """
    dirname = repo()
    try:
        os.makedirs(dirname)
        if util.via_sudo():
            uid = int(os.environ['SUDO_UID'])
            gid = int(os.environ['SUDO_GID'])
            os.chown(dirname, uid, gid)
    except OSError:
        pass
    try:
        p = subprocess.Popen(['git',
                              '--git-dir', dirname,
                              'init',
                              '--bare',
                              '-q'],
                             close_fds=True,
                             preexec_fn=unroot,
                             stdout=sys.stderr,
                             stderr=sys.stderr)
    except OSError:
        logging.error('git not found on PATH - exiting')
        sys.exit(1)
    p.communicate()
    if 0 != p.returncode:
        #sys.exit(p.returncode)
        raise GitError(p.returncode)


def git_args():
    """
    Return the basic arguments for running Git commands against
    the blueprints repository.
    """
    return ['git', '--git-dir', repo(), '--work-tree', os.getcwd()]


def git(*args, **kwargs):
    """
    Execute a Git command.  Raises GitError on non-zero exits unless the
    raise_exc keyword argument is falsey.
    """
    try:
        p = subprocess.Popen(git_args() + list(args),
                             close_fds=True,
                             preexec_fn=unroot,
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE)
    except OSError:
        logging.error('git not found on PATH - exiting')
        sys.exit(1)
    stdout, stderr = p.communicate(kwargs.get('stdin'))
    if 0 != p.returncode and kwargs.get('raise_exc', True):
        raise GitError(p.returncode)
    return p.returncode, stdout


def repo():
    """
    Return the full path to the Git repository.
    """
    return os.path.expandvars('$HOME/.blueprints.git')


def rev_parse(refname):
    """
    Return the referenced commit or None.
    """
    status, stdout = git('rev-parse', '-q', '--verify', refname,
                         raise_exc=False)
    if 0 != status:
        return None
    return stdout.rstrip()


def tree(commit):
    """
    Return the tree in the given commit or None.
    """
    status, stdout = git('show', '--pretty=format:%T', commit)
    if 0 != status:
        return None
    return stdout[0:40]


def ls_tree(tree, dirname=[]):
    """
    Generate all the pathnames in the given tree.
    """
    status, stdout = git('ls-tree', tree)
    for line in stdout.splitlines():
        mode, type, sha, filename = line.split()
        if 'tree' == type:
            for entry in ls_tree(sha, dirname + [filename]):
                yield entry
        else:
            yield mode, type, sha, os.path.join(*dirname + [filename])


def blob(tree, pathname):
    """
    Return the SHA of the blob by the given name in the given tree.
    """
    for mode, type, sha, pathname2 in ls_tree(tree):
        if pathname == pathname2:
            return sha
    return None


def content(blob):
    """
    Return the content of the given blob.
    """
    status, stdout = git('show', blob)
    if 0 != status:
        return None
    return stdout


def cat_file(blob, pathname=None):
    """
    If `pathname` is `None`, return an open file handle to the blob in
    Git's object store, otherwise stream the blob to `pathname`, all via
    the git-cat-file(1) command.
    """
    args = git_args() + ['cat-file', 'blob', blob]
    if pathname is None:
        return subprocess.Popen(args,
                                close_fds=True,
                                preexec_fn=unroot,
                                stdout=subprocess.PIPE).stdout
    else:
        subprocess.Popen(args,
                         close_fds=True,
                         preexec_fn=unroot,
                         stdout=open(pathname, 'w')).communicate()


def write_tree():
    status, stdout = git('write-tree')
    if 0 != status:
        return None
    return stdout.rstrip()


def commit_tree(tree, message='', parent=None):
    if parent is None:
        status, stdout = git('commit-tree', tree, stdin=message)
    else:
        status, stdout = git('commit-tree', tree, '-p', parent, stdin=message)
    if 0 != status:
        return None
    return stdout.rstrip()


def configured():
    """
    Return `True` if the author is configured in Git.  This allows Blueprint
    to bail out early for users that don't have things configured just right.
    """
    return not git('config', 'user.name', raise_exc=False)[0] \
       and not git('config', 'user.email', raise_exc=False)[0]



"""
Search for software built from source to include in the blueprint as a tarball.
"""

import errno
import glob
import hashlib
import logging
import os
import os.path
import re
import shutil
import stat
import subprocess
import tarfile

from blueprint import context_managers
from blueprint import util


def _source(b, r, dirname, old_cwd):
    tmpname = os.path.join(os.getcwd(), dirname[1:].replace('/', '-'))

    exclude = []

    pattern_pip = re.compile(r'\.egg-info/installed-files.txt$')
    pattern_egg = re.compile(r'\.egg(?:-info)?(?:/|$)')
    pattern_pth = re.compile(
        r'lib/python[^/]+/(?:dist|site)-packages/easy-install.pth$')
    pattern_bin = re.compile(
        r'EASY-INSTALL(?:-ENTRY)?-SCRIPT|This file was generated by RubyGems')

    # Create a partial shallow copy of the directory.
    for dirpath, dirnames, filenames in os.walk(dirname):

        # Definitely ignore the shallow copy directory.
        if dirpath.startswith(tmpname):
            continue

        # Determine if this entire directory should be ignored by default.
        ignored = r.ignore_file(dirpath)

        dirpath2 = os.path.normpath(
            os.path.join(tmpname, os.path.relpath(dirpath, dirname)))

        # Create this directory in the shallow copy with matching mode, owner,
        # and owning group.  Suggest running as `root` if this doesn't work.
        os.mkdir(dirpath2)
        s = os.lstat(dirpath)
        try:
            try:
                os.lchown(dirpath2, s.st_uid, s.st_gid)
            except OverflowError:
                logging.warning('{0} has uid:gid {1}:{2} - using chown(1)'.
                                format(dirpath, s.st_uid, s.st_gid))
                p = subprocess.Popen(['chown',
                                      '{0}:{1}'.format(s.st_uid, s.st_gid),
                                      dirpath2],
                                     close_fds=True)
                p.communicate()
            os.chmod(dirpath2, s.st_mode)
        except OSError as e:
            logging.warning('{0} caused {1} - try running as root'.
                            format(dirpath, errno.errorcode[e.errno]))
            return

        for filename in filenames:
            pathname = os.path.join(dirpath, filename)

            if r.ignore_source(pathname, ignored):
                continue

            pathname2 = os.path.join(dirpath2, filename)

            # Exclude files that are part of the RubyGems package.
            for globname in (
                os.path.join('/usr/lib/ruby/gems/*/gems/rubygems-update-*/lib',
                             pathname[1:]),
                os.path.join('/var/lib/gems/*/gems/rubygems-update-*/lib',
                             pathname[1:])):
                if 0 < len(glob.glob(globname)):
                    continue

            # Remember the path to all of `pip`'s `installed_files.txt` files.
            if pattern_pip.search(pathname):
                exclude.extend([os.path.join(dirpath2, line.rstrip())
                    for line in open(pathname)])

            # Likewise remember the path to Python eggs.
            if pattern_egg.search(pathname):
                exclude.append(pathname2)

            # Exclude `easy_install`'s bookkeeping file, too.
            if pattern_pth.search(pathname):
                continue

            # Exclude executable placed by Python packages or RubyGems.
            if pathname.startswith('/usr/local/bin/'):
                try:
                    if pattern_bin.search(open(pathname).read()):
                        continue
                except IOError as e:
                    pass

            # Exclude share/applications/mimeinfo.cache, whatever that is.
            if '/usr/local/share/applications/mimeinfo.cache' == pathname:
                continue

            # Clean up dangling symbolic links.  This makes the assumption
            # that no one intends to leave dangling symbolic links hanging
            # around, which I think is a good assumption.
            s = os.lstat(pathname)
            if stat.S_ISLNK(s.st_mode):
                try:
                    os.stat(pathname)
                except OSError as e:
                    if errno.ENOENT == e.errno:
                        logging.warning('ignored dangling symbolic link {0}'.
                                        format(pathname))
                        continue

            # Hard link this file into the shallow copy.  Suggest running as
            # `root` if this doesn't work though in practice the check above
            # will have already caught this problem.
            try:
                os.link(pathname, pathname2)
            except OSError as e:
                logging.warning('{0} caused {1} - try running as root'.
                                format(pathname, errno.errorcode[e.errno]))
                return

    # Unlink files that were remembered for exclusion above.
    for pathname in exclude:
        try:
            os.unlink(pathname)
        except OSError as e:
            if e.errno not in (errno.EISDIR, errno.ENOENT):
                raise e

    # Remove empty directories.  For any that hang around, match their
    # access and modification times to the source, otherwise the hash of
    # the tarball will not be deterministic.
    for dirpath, dirnames, filenames in os.walk(tmpname, topdown=False):
        try:
            os.rmdir(dirpath)
        except OSError:
            s = os.lstat(os.path.join(dirname, os.path.relpath(dirpath,
                                                               tmpname)))
            os.utime(dirpath, (s.st_atime, s.st_mtime))

    # If the shallow copy of still exists, create a tarball named by its
    # SHA1 sum and include it in the blueprint.
    try:
        tar = tarfile.open('tmp.tar', 'w')
        tar.add(tmpname, '.')
    except OSError:
        return
    finally:
        tar.close()
    sha1 = hashlib.sha1()
    f = open('tmp.tar', 'r')
    [sha1.update(buf) for buf in iter(lambda: f.read(4096), '')]
    f.close()
    tarname = '{0}.tar'.format(sha1.hexdigest())
    shutil.move('tmp.tar', os.path.join(old_cwd, tarname))
    b.add_source(dirname, tarname)


def sources(b, r):
    logging.info('searching for software built from source')
    for pathname, negate in r['source']:
        if negate and os.path.isdir(pathname) \
        and not r.ignore_source(pathname):

            # Note before creating a working directory within pathname what
            # it's atime and mtime should be.
            s = os.lstat(pathname)

            # Create a working directory within pathname to avoid potential
            # EXDEV when creating the shallow copy and tarball.
            try:
                with context_managers.mkdtemp(pathname) as c:

                    # Restore the parent of the working directory to its
                    # original atime and mtime, as if pretending the working
                    # directory never actually existed.
                    os.utime(pathname, (s.st_atime, s.st_mtime))

                    # Create the shallow copy and possibly tarball of the
                    # relevant parts of pathname.
                    _source(b, r, pathname, c.cwd)

                # Once more restore the atime and mtime after the working
                # directory is destroyed.
                os.utime(pathname, (s.st_atime, s.st_mtime))

            # If creating the temporary directory fails, bail with a warning.
            except OSError as e:
                logging.warning('{0} caused {1} - try running as root'.
                                format(pathname, errno.errorcode[e.errno]))

    if 0 < len(b.sources):
        b.arch = util.arch()



"""
Testing out Librato's metrics platform.
"""

from ConfigParser import NoOptionError, NoSectionError
import base64
import httplib
import urllib

from blueprint import cfg


try:
    token = cfg.get('librato', 'token')
    username = cfg.get('librato', 'username')
    auth = 'Basic {0}'.format(base64.b64encode('{0}:{1}'.format(username,
                                                                token)))
except (NoOptionError, NoSectionError):
    auth = None


def count(name, value=1):
    """
    Update a counter in Librato's metrics platform.
    """
    if auth is None:
        return
    conn = httplib.HTTPSConnection('metrics-api.librato.com')
    conn.request('POST',
                 '/v1/counters/{0}.json'.format(urllib.quote(name)),
                 urllib.urlencode({'value': value}),
                 {'Authorization': auth,
                  'Content-Type': 'application/x-www-form-urlencoded'})
    r = conn.getresponse()
    conn.close()



"""
Search for PEAR/PECL packages to include in the blueprint.
"""

import logging
import re
import subprocess

from blueprint import util


def php(b, r):
    logging.info('searching for PEAR/PECL packages')

    # Precompile a pattern for parsing the output of `{pear,pecl} list`.
    pattern = re.compile(r'^([0-9a-zA-Z_]+)\s+([0-9][0-9a-zA-Z\.-]*)\s')

    # PEAR packages are managed by `php-pear` (obviously).  PECL packages
    # are managed by `php5-dev` because they require development headers
    # (less obvious but still makes sense).
    if util.lsb_release_codename() is None:
        pecl_manager = 'php-devel'
    else:
        pecl_manager = 'php5-dev'
    for manager, progname in (('php-pear', 'pear'),
                              (pecl_manager, 'pecl')):

        try:
            p = subprocess.Popen([progname, 'list'],
                                 close_fds=True, stdout=subprocess.PIPE)
        except OSError:
            continue
        for line in p.stdout:
            match = pattern.match(line)
            if match is None:
                continue
            package, version = match.group(1), match.group(2)
            if not r.ignore_package(manager, package):
                b.add_package(manager, package, version)



"""
Managers are callable strings that can generate the commands needed by
resources.  They're mostly useful in the context of generated shell code.
"""

import re


class PackageManager(unicode):
    """
    Package managers each have their own syntax.  All supported package
    managers are encapsulated in this manager class.
    """

    def gate(self, package, version, relaxed=False):
        """
        Return a shell command that checks for the given version of the
        given package via this package manager.  It should exit non-zero
        if the package is not in the desired state.
        """
        if version is None:
            relaxed = True

        if 'apt' == self:
            if relaxed:
                return 'dpkg-query -W {0} >/dev/null'.format(package)
            else:
                return ('[ "$(dpkg-query -f\'${{{{Version}}}}\\n\' -W {0})" '
                        '= "{1}" ]').format(package, version)

        if 'rpm' == self:
            return 'rpm -q {0} >/dev/null'.format(package)
        if 'yum' == self:
            if relaxed:
                arg = package
            else:
                match = re.match(r'^\d+:(.+)$', version)
                if match is None:
                    arg = '{0}-{1}'.format(package, version)
                else:
                    arg = '{0}-{1}'.format(package, match.group(1))
            return 'rpm -q {0} >/dev/null'.format(arg)

        if 'rubygems' == self:
            if relaxed:
                return 'gem list -i {0} >/dev/null'.format(package)
            else:
                return 'gem -i -v{1} {0} >/dev/null'.format(package, version)
        match = re.match(r'^ruby(?:gems)?(\d+\.\d+(?:\.\d+)?)', self)
        if match is not None:
            if relaxed:
                return ('gem{0} list -i {1} >/dev/null'.
                        format(match.group(1), package))
            else:
                return ('gem{0} list -i -v{2} {1} >/dev/null'.
                        format(match.group(1), package, version))

        return None

    def install(self, package, version, relaxed=False):
        """
        Return a shell command that installs the given version of the given
        package via this package manager.
        """
        if version is None:
            relaxed = True

        if 'apt' == self:
            arg = package if relaxed else '{0}={1}'.format(package, version)
            return ('apt-get -y -q -o DPkg::Options::=--force-confold '
                    'install {0}').format(arg)

        # AWS cfn-init templates may specify RPMs to be installed from URLs,
        # which are specified as versions.
        if 'rpm' == self:
            return 'rpm -U {0}'.format(version)

        if 'yum' == self:
            arg = package if relaxed else '{0}-{1}'.format(package, version)
            return 'yum -y install {0}'.format(arg)

        if 'rubygems' == self:
            if relaxed:
                return 'gem install --no-rdoc --no-ri {0}'.format(package)
            else:
                return ('gem install --no-rdoc --no-ri -v{1} {0}'.
                        format(package, version))
        match = re.match(r'^ruby(?:gems)?(\d+\.\d+(?:\.\d+)?)', self)
        if match is not None:
            if relaxed:
                return ('gem{0} install --no-rdoc --no-ri {1}'.
                        format(match.group(1), package))
            else:
                return ('gem{0} install --no-rdoc --no-ri -v{2} {1}'.
                        format(match.group(1), package, version))

        if 'python' == self:
            return 'easy_install {0}'.format(package)
        match = re.match(r'^python(\d+\.\d+)', self)
        if match is not None:
            return 'easy_install-{0} {1}'.format(match.group(1), package)
        if 'pip' == self or 'python-pip' == self:
            arg = package if relaxed else '{0}=={1}'.format(package, version)
            return 'pip install {0}'.format(arg)

        if 'php-pear' == self:
            arg = package if relaxed else '{0}-{1}'.format(package, version)
            return 'pear install {0}'.format(arg)
        if self in ('php5-dev', 'php-devel'):
            arg = package if relaxed else '{0}-{1}'.format(package, version)
            return 'pecl install {0}'.format(arg)

        if 'nodejs' == self:
            arg = package if relaxed else '{0}@{1}'.format(package, version)
            return 'npm install -g {0}'.format(arg)

        if relaxed:
            return ': unknown manager {0} for {1}'.format(self, package)
        else:
            return ': unknown manager {0} for {1} {2}'.format(self,
                                                              package,
                                                              version)

    def __call__(self, package, version, relaxed=False):
        """
        Return a shell command that checks for and possibly installs
        the given version of the given package.
        """
        gate = self.gate(package, version, relaxed)
        install = self.install(package, version, relaxed)
        if gate is None:
            return install
        return gate + ' || ' + install


class ServiceManager(unicode):
    """
    Service managers each have their own syntax.  All supported service
    managers are encapsulated in this manager class.
    """

    _env_pattern = re.compile(r'[^0-9A-Za-z]')

    def env_var(self, service):
        """
        Return the name of the environment variable being used to track the
        state of the given service.
        """
        return 'SERVICE_{0}_{1}'.format(self._env_pattern.sub('', self),
                                        self._env_pattern.sub('', service))

    def __call__(self, service):
        """
        Return a shell command that restarts the given service via this
        service manager.
        """

        if 'upstart' == self:
            return ('[ -n "${0}" ] && {{{{ restart {1} || start {1}; }}}}'.
                format(self.env_var(service), service))

        return '[ -n "${0}" ] && /etc/init.d/{1} restart'.format(
            self.env_var(service), service)



"""
Instantiate Blueprint objects for the command-line tools.  Use of these
functions outside of command-line tools is not advised, as in many cases
they exit the Python interpreter.
"""

import logging
import os
import sys

import blueprint
import context_managers
import rules


def create(options, args):
    """
    Instantiate and return a Blueprint object from either standard input or by
    reverse-engineering the system.
    """
    try:
        with context_managers.mkdtemp():

            if not os.isatty(sys.stdin.fileno()):
                try:
                    b = blueprint.Blueprint.load(sys.stdin, args[0])
                except ValueError:
                    logging.error(
                        'standard input contains invalid blueprint JSON')
                    sys.exit(1)
            else:
                b = blueprint.Blueprint.create(args[0])

            if options.subtrahend:
                logging.info('subtracting {0}'.format(options.subtrahend))
                b_s = blueprint.Blueprint.checkout(options.subtrahend)
                b = b - b_s

            b.commit(options.message or '')
            return b

    except blueprint.NameError:
        logging.error('invalid blueprint name')
        sys.exit(1)


def read(options, args):
    """
    Instantiate and return a Blueprint object from either standard input or by
    reading from the local Git repository.
    """
    try:
        name = args[0]
    except IndexError:
        name = None
    name, stdin = '-' == name and (None, True) or (name, False)
    try:
        if not os.isatty(sys.stdin.fileno()) or stdin:
            try:

                # TODO This implementation won't be able to find source
                # tarballs that should be associated with the blueprint
                # on standard input.
                return blueprint.Blueprint.load(sys.stdin, name)

            except ValueError:
                logging.error('standard input contains invalid blueprint JSON')
                sys.exit(1)
        if name is not None:
            return blueprint.Blueprint.checkout(name)
    except blueprint.NotFoundError:
        logging.error('blueprint {0} does not exist'.format(name))
        sys.exit(1)
    except blueprint.NameError:
        logging.error('invalid blueprint name {0}'.format(name))
        sys.exit(1)
    logging.error('no blueprint found on standard input')
    sys.exit(1)


def read_rules(options, args):
    """
    Instantiate and return a Blueprint object created by rules read from
    either standard input or the given pathname.
    """
    try:
        pathname = args[0]
    except IndexError:
        pathname = None
    pathname, stdin = '-' == pathname and (None, True) or (pathname, False)
    r = rules.none()
    if not os.isatty(sys.stdin.fileno()) or stdin:
        r.parse(sys.stdin)
        with context_managers.mkdtemp():
            b = blueprint.Blueprint.rules(r, 'blueprint-rendered-rules')
            b.commit(options.message or '')
            return b
    if pathname is not None:
        name, _ = os.path.splitext(os.path.basename(pathname))
        try:
            r.parse(open(pathname))
            with context_managers.mkdtemp():
                b = blueprint.Blueprint.rules(r, name)
                b.commit(options.message or '')
                return b
        except blueprint.NameError:
            logging.error('invalid blueprint name {0}'.format(name))
            sys.exit(1)
        except IOError:
            logging.error('{0} does not exist'.format(pathname))
            sys.exit(1)
    logging.error('no rules found on standard input')
    sys.exit(1)



"""
Every backend provider defines a function named as the module that contains
it, which will be called when a blueprint is created.  This module gathers
those functions together.
"""

import glob
import os.path
import sys


__all__ = [os.path.basename(filename)[:-3]
           for filename in glob.glob(os.path.join(os.path.dirname(__file__),
                                                  '[!_]*.py'))]
for name in __all__:
    module = __import__(name, globals(), locals(), [], 1)
    setattr(sys.modules[__name__], name, getattr(module, name))



"""
Search for Ruby gems to include in the blueprint.
"""

import glob
import logging
import os
import re

from blueprint import util


def gem(b, r):
    logging.info('searching for Ruby gems')

    # Precompile a pattern for extracting the version of Ruby that was used
    # to install the gem.
    pattern = re.compile(r'gems/([^/]+)/gems')

    # Look for gems in all the typical places.  This is easier than looking
    # for `gem` commands, which may or may not be on `PATH`.
    for globname in ('/usr/lib/ruby/gems/*/gems',
                     '/usr/local/lib/ruby/gems/*/gems',
                     '/var/lib/gems/*/gems'):
        for dirname in glob.glob(globname):

            # The `ruby1.9.1` (really 1.9.2) package on Maverick begins
            # including RubyGems in the `ruby1.9.1` package and marks the
            # `rubygems1.9.1` package as virtual.  So for Maverick and
            # newer, the manager is actually `ruby1.9.1`.
            match = pattern.search(dirname)
            if '1.9.1' == match.group(1) and util.rubygems_virtual():
                manager = 'ruby{0}'.format(match.group(1))

            # Oneiric and RPM-based distros just have one RubyGems package.
            elif util.rubygems_unversioned():
                manager = 'rubygems'

            # Debian-based distros qualify the package name with the version
            # of Ruby it will use.
            else:
                manager = 'rubygems{0}'.format(match.group(1))

            for entry in os.listdir(dirname):
                try:
                    package, version = entry.rsplit('-', 1)
                except ValueError:
                    logging.warning('skipping questionably named gem {0}'.
                                    format(entry))
                    continue
                if not r.ignore_package(manager, package):
                    b.add_package(manager, package, version)



"""
Every frontend provider defines a function named as the module that contains
it, which is to be used as its main entrypoint.  Only one frontend is
typically in use so, unlike with backends, they're not preloaded.
"""



"""
Utility functions.
"""

import json
import os
import os.path
import re
import subprocess


def arch():
    """
    Return the system's architecture according to dpkg or rpm.
    """
    try:
        p = subprocess.Popen(['dpkg', '--print-architecture'],
                             close_fds=True, stdout=subprocess.PIPE)
    except OSError as e:
        p = subprocess.Popen(['rpm', '--eval', '%_arch'],
                             close_fds=True, stdout=subprocess.PIPE)
    stdout, stderr = p.communicate()
    if 0 != p.returncode:
        return None
    return stdout.rstrip()


def lsb_release_codename():
    """
    Return the OS release's codename.
    """
    if hasattr(lsb_release_codename, '_cache'):
        return lsb_release_codename._cache
    try:
        p = subprocess.Popen(['lsb_release', '-c'], stdout=subprocess.PIPE)
    except OSError:
        lsb_release_codename._cache = None
        return lsb_release_codename._cache
    stdout, stderr = p.communicate()
    if 0 != p.returncode:
        lsb_release_codename._cache = None
        return lsb_release_codename._cache
    match = re.search(r'\t(\w+)$', stdout)
    if match is None:
        lsb_release_codename._cache = None
        return lsb_release_codename._cache
    lsb_release_codename._cache = match.group(1)
    return lsb_release_codename._cache


# Patterns for determining which Upstart services should be included, based
# on the events used to start them.
pattern_upstart_1 = re.compile(r'start\s+on\s+runlevel\s+\[[2345]', re.S)
pattern_upstart_2 = re.compile(r'start\s+on\s+\([^)]*(?:filesystem|filesystems|local-filesystems|mounted|net-device-up|remote-filesystems|startup|virtual-filesystems)[^)]*\)', re.S)


def parse_service(pathname):
    """
    Parse a potential service init script or config file into the
    manager and service name or raise `ValueError`.  Use the Upstart
    "start on" stanzas and SysV init's LSB headers to restrict services to
    only those that start at boot and run all the time.
    """
    dirname, basename = os.path.split(pathname)
    if '/etc/init' == dirname:
        service, ext = os.path.splitext(basename)

        # Ignore extraneous files in /etc/init.
        if '.conf' != ext:
            raise ValueError('not an Upstart config')

        # Ignore services that don't operate on the (faked) main runlevels.
        try:
            content = open(pathname).read()
        except IOError:
            raise ValueError('not a readable Upstart config')
        if not (pattern_upstart_1.search(content) \
                or pattern_upstart_2.search(content)):
            raise ValueError('not a running service')

        return ('upstart', service)
    elif '/etc/init.d' == dirname or '/etc/rc.d/init.d' == dirname:

        # Let Upstart handle its services.
        if os.path.islink(pathname) \
            and '/lib/init/upstart-job' == os.readlink(pathname):
            raise ValueError('proxy for an Upstart config')

        # Ignore services that don't operate on the main runlevels.
        try:
            content = open(pathname).read()
        except IOError:
            raise ValueError('not a readable SysV init script')
        if not re.search(r'(?:Default-Start|chkconfig):\s*[2345]', content):
            raise ValueError('not a running service')

        return ('sysvinit', basename)
    else:
        raise ValueError('not a service')


def rubygems_unversioned():
    """
    Determine whether RubyGems is suffixed by the Ruby language version.
    It ceased to be on Oneiric.  It always has been on RPM-based distros.
    """
    codename = lsb_release_codename()
    return codename is None or codename[0] >= 'o'


def rubygems_update():
    """
    Determine whether the `rubygems-update` gem is needed.  It is needed
    on Lucid and older systems.
    """
    codename = lsb_release_codename()
    return codename is not None and codename[0] < 'm'


def rubygems_virtual():
    """
    Determine whether RubyGems is baked into the Ruby 1.9 distribution.
    It is on Maverick and newer systems.
    """
    codename = lsb_release_codename()
    return codename is not None and codename[0] >= 'm'


def rubygems_path():
    """
    Determine based on the OS release where RubyGems will install gems.
    """
    if lsb_release_codename() is None or rubygems_update():
        return '/usr/lib/ruby/gems'
    return '/var/lib/gems'


def via_sudo():
    """
    Return `True` if Blueprint was invoked via `sudo`(8), which indicates
    that privileges must be dropped when writing to the filesystem.
    """
    return 'SUDO_UID' in os.environ \
        and 'SUDO_GID' in os.environ \
        and 'blueprint' in os.environ.get('SUDO_COMMAND', '')


class BareString(unicode):
    """
    Strings of this type will not be quoted when written into a Puppet
    manifest or Chef cookbook.
    """
    pass


class JSONEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, set):
            return list(o)
        return super(JSONEncoder, self).default(o)

def json_dumps(o):
    return JSONEncoder(indent=2, sort_keys=True).encode(o)


def unicodeme(s):
    if isinstance(s, unicode):
        return s
    for encoding in ('utf_8', 'latin_1'):
        try:
            return unicode(s, encoding)
        except UnicodeDecodeError:
            pass
    # TODO Issue a warning?
    return s



import os
import shutil
import tempfile

from blueprint import util


class cd(object):
    """
    Run in an alternative working directory in this context.
    """

    def __init__(self, new_cwd):
        self.new_cwd = new_cwd

    def __enter__(self):
        self.old_cwd = os.getcwd()
        os.chdir(self.new_cwd)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        os.chdir(self.old_cwd)


class mkdtemp(object):
    """
    Run in a temporary working directory in this context.  Remove the
    temporary directory automatically afterward.
    """

    def __init__(self, dir=None):
        self.cwd = os.getcwd()
        if dir is None:
            dir = tempfile.gettempdir()
        self.tempdir = tempfile.mkdtemp(dir=dir)
        if util.via_sudo():
            uid = int(os.environ['SUDO_UID'])
            gid = int(os.environ['SUDO_GID'])
            os.chown(self.tempdir, uid, gid)

    def __enter__(self):
        os.chdir(self.tempdir)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        os.chdir(self.cwd)
        shutil.rmtree(self.tempdir)



"""
Shell code generator.
"""

import codecs
from collections import defaultdict
import gzip as gziplib
import os
import os.path
import re
from shutil import copyfile
import tarfile
import unicodedata

from blueprint import git
from blueprint import util


def sh(b, relaxed=False, server='https://devstructure.com', secret=None):
    """
    Generate shell code.
    """
    s = Script(b.name, comment=b.DISCLAIMER)

    # Build an inverted index (lookup table, like in hardware, hence the name)
    # of service dependencies to services.
    lut = {'files': defaultdict(set),
           'packages': defaultdict(lambda: defaultdict(set)),
           'sources': defaultdict(set)}
    def service_file(manager, service, pathname):
        lut['files'][pathname].add((manager, service))
    def service_package(manager, service, package_manager, package):
        lut['packages'][package_manager][package].add((manager, service))
    def service_source(manager, service, dirname):
        lut['sources'][dirname].add((manager, service))
    b.walk(service_file=service_file,
           service_package=service_package,
           service_source=service_source)

    commit = git.rev_parse(b.name)
    tree = None if commit is None else git.tree(commit)
    def source(dirname, filename, gen_content, url):
        """
        Extract a source tarball.
        """
        if dirname in lut['sources']:
            s.add('MD5SUM="$(find "{0}" -printf %T@\\\\n | md5sum)"',
                  args=(dirname,))
        if url is not None:
            s.add_list(('curl -o "{0}" "{1}"',),
                       ('wget -O "{0}" "{1}"',),
                       args=(filename, url),
                       operator='||')
            if '.zip' == pathname[-4:]:
                s.add('unzip "{0}" -d "{1}"', args=(filename, dirname))
            else:
                s.add('mkdir -p "{1}" && tar xf "{0}" -C "{1}"', args=(filename, dirname))
        elif secret is not None:
            s.add_list(('curl -O "{0}/{1}/{2}/{3}"',),
                       ('wget "{0}/{1}/{2}/{3}"',),
                       args=(server, secret, b.name, filename),
                       operator='||')
            s.add('mkdir -p "{1}" && tar xf "{0}" -C "{1}"', args=(filename, dirname))
        elif gen_content is not None:
            s.add('mkdir -p "{1}" && tar xf "{0}" -C "{1}"', args=(filename, dirname))
            s.add_source(filename, git.blob(tree, filename))
        for manager, service in lut['sources'][dirname]:
            s.add_list(('[ "$MD5SUM" != "$(find "{0}" -printf %T@\\\\n '
                        '| md5sum)" ]',),
                       ('{1}=1',),
                       args=(dirname, manager.env_var(service)),
                       operator='&&')

    def file(pathname, f):
        """
        Place a file.
        """
        if pathname in lut['files']:
            s.add('MD5SUM="$(md5sum "{0}" 2>/dev/null)"', args=(pathname,))
        s.add('mkdir -p "{0}"', args=(os.path.dirname(pathname),))
        if '120000' == f['mode'] or '120777' == f['mode']:
            s.add('ln -s "{0}" "{1}"', args=(f['content'], pathname))
        else:
            if 'source' in f:
                s.add_list(('curl -o "{0}" "{1}"',),
                           ('wget -O "{0}" "{1}"',),
                           args=(pathname, f['source']),
                           operator='||')
            else:
                if 'template' in f:
                    s.templates = True
                    if 'base64' == f['encoding']:
                        commands = ('base64 --decode', 'mustache')
                    else:
                        commands = ('mustache',)
                    s.add_list(('set +x',),
                               ('. "lib/mustache.sh"',),
                               ('for F in */blueprint-template.d/*.sh',),
                               ('do',),
                               ('\t. "$F"',),
                               ('done',),
                               (f.get('data', '').rstrip(),),
                               (command(*commands,
                                        escape_stdin=True,
                                        stdin=f['template'],
                                        stdout=pathname),),
                               operator='\n',
                               wrapper='()')
                else:
                    if 'base64' == f['encoding']:
                        commands = ('base64 --decode',)
                    else:
                        commands = ('cat',)
                    s.add(*commands, stdin=f['content'], stdout=pathname)
            if 'root' != f['owner']:
                s.add('chown {0} "{1}"', args=(f['owner'], pathname))
            if 'root' != f['group']:
                s.add('chgrp {0} "{1}"', args=(f['group'], pathname))
            if '100644' != f['mode']:
                s.add('chmod {0} "{1}"', args=(f['mode'][-4:], pathname))
        for manager, service in lut['files'][pathname]:
            s.add('[ "$MD5SUM" != "$(md5sum "{0}")" ] && {1}=1',
                  args=(pathname, manager.env_var(service)))

    def before_packages(manager):
        """
        Configure the package managers.
        """
        if manager not in b.packages:
            return
        if 'apt' == manager:
            s.add('export APT_LISTBUGS_FRONTEND="none"')
            s.add('export APT_LISTCHANGES_FRONTEND="none"')
            s.add('export DEBIAN_FRONTEND="noninteractive"')
            s.add('apt-get -q update')
        elif 'yum' == manager:
            s.add('yum makecache')

    def package(manager, package, version):
        """
        Install a package.
        """
        if manager == package:
            return

        if manager in lut['packages'] and package in lut['packages'][manager]:
            s.add_list((manager.gate(package, version, relaxed),),
                       (command_list((manager.install(package,
                                                      version,
                                                      relaxed),),
                                     *[('{0}=1'.format(m.env_var(service)),)
                                       for m, service in
                                       lut['packages'][manager][package]],
                                     wrapper='{{}}'),),
                       operator='||')
        else:
            s.add(manager(package, version, relaxed))

        if manager not in ('apt', 'rpm', 'yum'):
            return

        # See comments on this section in `blueprint.frontend.puppet`.
        match = re.match(r'^rubygems(\d+\.\d+(?:\.\d+)?)$', package)
        if match is not None and util.rubygems_update():
            s.add('/usr/bin/gem{0} install --no-rdoc --no-ri rubygems-update',
                  args=(match.group(1),))
            s.add('/usr/bin/ruby{0} $(PATH=$PATH:/var/lib/gems/{0}/bin '
                  'which update_rubygems)',
                  args=(match.group(1),))

        if 'nodejs' == package:
            s.add_list(('which npm',),
                       (command_list(('curl http://npmjs.org/install.sh',),
                                     ('wget -O- http://npmjs.org/install.sh',),
                                     operator='||',
                                     wrapper='{{}}'),
                        'sh'),
                operator='||')

    def service(manager, service):
        s.add(manager(service))

    b.walk(source=source,
           file=file,
           before_packages=before_packages,
           package=package,
           service=service)

    return s


def command(*commands, **kwargs):
    commands = list(commands)
    if 'stdout' in kwargs:
        commands[-1] += ' >"{0}"'.format(kwargs['stdout'])
    if 'stdin' in kwargs:
        stdin = (kwargs['stdin'].replace(u'\\', u'\\\\').
                                 replace(u'$', u'\\$').
                                 replace(u'`', u'\\`'))
        if kwargs.get('escape_stdin', False):
            stdin = stdin.replace(u'{', u'{{').replace(u'}', u'}}')
        eof = 'EOF'
        while eof in stdin:
            eof += 'EOF'
        commands[0] += ' <<{0}'.format(eof)
        return ''.join([' | '.join(commands).format(*kwargs.get('args', ())),
                        '\n',
                        stdin,
                        '' if '' == stdin or '\n' == stdin[-1] else '\n',
                        eof])
    return ' | '.join(commands).format(*kwargs.get('args', ()))


def command_list(*commands, **kwargs):
    operator = {'&&': u' && ',
                '||': u' || ',
                '\n': u'\n',
                ';': u'; '}[kwargs.get('operator', ';')]
    wrapper = {'()': (u'(\n', u'\n)') if u'\n' == operator else (u'(', u')'),
               '{}': (u'{ ', u'; }'),
               '{{}}': (u'{{ ', u'; }}'), # Prevent double-escaping.
               '': (u'', u'')}[kwargs.get('wrapper', '')]
    return wrapper[0] \
         + operator.join([command(*c, **kwargs) for c in commands]) \
         + wrapper[-1]


class Script(object):
    """
    A script is a list of shell commands.  The pomp and circumstance is
    only necessary for providing an interface like the Puppet and Chef
    code generators.
    """

    def __init__(self, name, comment=None):
        if name is None:
            self.name = 'blueprint-generated-shell-script'
        else:
            self.name = name
        self.out = [comment if comment is not None else '',
                    'set -x\n',
                    'cd "$(dirname "$0")"\n']
        self.sources = {}
        self.templates = False

    def add(self, s='', *args, **kwargs):
        self.out.append((unicode(s) + u'\n').format(*args))
        for filename, content in kwargs.get('sources', {}).iteritems():
            self.sources[filename] = content

    def add(self, *args, **kwargs):
        """
        Add a command or pipeline to the `Script`.  Each positional `str`
        is an element in the pipeline.  The keyword argument `args`, if
        present, should contain an iterable of arguments to be substituted
        into the final pipeline by the new-style string formatting library.
        """
        self.out.append(command(*args, **kwargs))

    def add_list(self, *args, **kwargs):
        """
        Add a command or pipeline, or list of commands or pipelines, to
        the `Script`.  Each positional `str` or `tuple` argument is a
        pipeline.  The keyword argument `operator`, if present, must be
        `';'`, `'&&'`, or `'||'` to control how the pipelines are joined.
        The keyword argument `stdin`, if present, should contain a string
        that will be given heredoc-style.  The keyword argument `stdout`,
        if present, should contain a string pathname that will receive
        standard output.  The keyword argument `args`, if present, should
        contain an iterable of arguments to be substituted into the final
        pipeline by the new-style string formatting library.
        """
        self.out.append(command_list(*args, **kwargs))

    def add_source(self, filename, blob):
        """
        Add a reference to a source tarball to the `Script`.  It will be
        placed in the output directory/tarball later via `git-cat-file`(1).
        """
        self.sources[filename] = blob

    def dumps(self):
        """
        Generate a string containing shell code and all file contents.
        """
        return ''.join(self.out)

    def dumpf(self, gzip=False):
        """
        Generate a file containing shell code and all file contents.
        """

        # Open a file by the correct name, possibly with inline gzipping.
        if 0 < len(self.sources) or self.templates:
            os.mkdir(self.name)
            filename = os.path.join(self.name, 'bootstrap.sh')
            f = codecs.open(filename, 'w', encoding='utf-8')
        elif gzip:
            filename = '{0}.sh.gz'.format(self.name)
            f = gziplib.open(filename, 'w')
        else:
            filename = '{0}.sh'.format(self.name)
            f = codecs.open(filename, 'w', encoding='utf-8')

        # Bring along `mustache.sh`, the default template data files, and
        # any user-provided template data files.
        if self.templates:
            os.mkdir(os.path.join(self.name, 'etc'))
            os.mkdir(os.path.join(self.name, 'etc', 'blueprint-template.d'))
            os.mkdir(os.path.join(self.name, 'lib'))
            os.mkdir(os.path.join(self.name, 'lib', 'blueprint-template.d'))
            copyfile(os.path.join(os.path.dirname(__file__), 'mustache.sh'),
                     os.path.join(self.name, 'lib', 'mustache.sh'))
            for src, dest in [('/etc/blueprint-template.d', 'etc'),
                              (os.path.join(os.path.dirname(__file__),
                                            'blueprint-template.d'),
                               'lib')]:
                try:
                    for filename2 in os.listdir(src):
                        if filename2.endswith('.sh'):
                            copyfile(os.path.join(src, filename2),
                                     os.path.join(self.name,
                                                  dest,
                                                  'blueprint-template.d',
                                                  filename2))
                except OSError:
                    pass

        # Write the actual shell code.
        for out in self.out:
            if isinstance(out, unicode):
                out = unicodedata.normalize('NFKD', out).encode('utf-8', 'ignore')
            f.write('{0}\n'.format(out))
        f.close()

        # Bring source tarballs along.
        for filename2, blob in sorted(self.sources.iteritems()):
            git.cat_file(blob, os.path.join(self.name, filename2))

        # Possibly gzip the result.
        if gzip and (0 < len(self.sources) or self.templates):
            filename = 'sh-{0}.tar.gz'.format(self.name)
            tarball = tarfile.open(filename, 'w:gz')
            tarball.add(self.name)
            tarball.close()
            return filename

        return filename



"""
Search for npm packages to include in the blueprint.  This assumes that
Node itself is installed via Chris Lea's PPAs, either
<https://launchpad.net/~chris-lea/+archive/node.js> or
<https://launchpad.net/~chris-lea/+archive/node.js-devel>.
"""

import logging
import re
import subprocess


def npm(b, r):
    logging.info('searching for npm packages')

    # Precompile a pattern for parsing the output of `{pear,pecl} list`.
    pattern = re.compile(r'^\S+ (\S+)@(\S+)$')

    try:
        p = subprocess.Popen(['npm', 'ls', '-g'],
                             close_fds=True,
                             stdout=subprocess.PIPE)
        for line in p.stdout:
            match = pattern.match(line.rstrip())
            if match is None:
                continue
            package, version = match.group(1), match.group(2)
            if not r.ignore_package('nodejs', package):
                b.add_package('nodejs', package, version)
    except OSError:
        pass
